/*
ftpexecd (ftpexec daemon) - This that monitors a proftpd log fifo and 
   executes an external program/script after a file is uploaded
 
Copyright (C) 2003 by Thomas Lachmann <TL@bunker-werk.net>
 
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
 
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
 
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <fcntl.h>
#include <string.h>
#include <syslog.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <unistd.h>
#include <sys/wait.h>
#include <pwd.h>
#include <grp.h>
#include <signal.h>
#include <time.h>

#include "ftpexecd.h"
#include "privs.h"
#include "version.h"
#include "config.h"

int LogToFile(char *msg);
void safe_exit(int returncode);

/* formats the time string */
char *fmt_time(time_t t)
{
  static char buf[30];
  static char *mons[] =
    { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };
  static char *days[] =
    { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };
  struct tm *tr;

  memset(buf, '\0', sizeof(buf));
  if ((tr = localtime(&t)) != NULL)
    {
      snprintf(buf, sizeof(buf), "%s %s %2d %02d:%02d:%02d %d",
               days[tr->tm_wday],
               mons[tr->tm_mon],
               tr->tm_mday,
               tr->tm_hour, tr->tm_min, tr->tm_sec, tr->tm_year + 1900);
    }
  else
    buf[0] = '\0';
  buf[sizeof(buf) - 1] = '\0';

  return buf;
}


/* prints a msg to syslog */
void print_log(int priority, char *fmt, ...)
{
  char buff[MAXBUF];
  va_list ap;
  char msg[MAXBUF];

  va_start(ap, fmt);
  vsnprintf(buff, sizeof(buff) - 1, fmt, ap);
  switch (priority)
    {
    case LOG_ERR:
      sprintf(msg, "ERROR: %s", buff);
      break;
    case LOG_WARNING:
      sprintf(msg, "WARNING: %s", buff);
      break;
    case LOG_INFO:
      sprintf(msg, "INFO: %s", buff);
      break;
    case LOG_DEBUG:
      sprintf(msg, "DEBUG: %s", buff);
      break;
    default:
      sprintf(msg, buff);
    }


  if (writeLog)
    {
      if (syslog_flag)
        syslog(priority, msg);
      else if (LogToFile(msg) != 0)
        syslog(LOG_ERR, "failed to open logFile %s", logFile);
    }

  if ((nodaemon_flag) || (stderr != NULL))
    fprintf(stderr, "%s\n", msg);

  va_end(ap);
}


/* prints to logfile */
int LogToFile(char *msg)
{
  PRIVS_ROOT
  fileLog = fopen(logFile, "a+");
  PRIVS_RELINQUISH
  if (fileLog != NULL)
    {
      fprintf(fileLog, "%s: %s\n", fmt_time(time(NULL)), msg);
      fclose(fileLog);
      return (0);
    }
  else
    return (-1);
}


/* closeall() -- close all FDs >= a specified value */
void closeall(int fd)
{
  int fdlimit = sysconf(_SC_OPEN_MAX);

  while (fd < fdlimit)
    close(fd++);
}

/* checks if a file exists */
int file_exists(char *file_name)
{
  struct stat file_stat;
  return ((stat(file_name, &file_stat) == 0) ? 1 : 0);
}


/* exits the daemo in a safe way if an error encountered */
void safe_exit(int returncode)
{
  print_log(LOG_INFO, "shutdown ftpexecd");
  closeall(0);
  exit(returncode);
}

/* execute a command with regard to euid & egid */
int safe_exec(char *cmd)
{
  int status = 0;
  pid_t pid;
  struct sigaction sa_ignore, sa_intr, sa_quit;
  sigset_t set_chldmask, set_save;

  if (!cmd)
    return -1;

  sa_ignore.sa_handler = SIG_IGN;
  sigemptyset(&sa_ignore.sa_mask);
  sa_ignore.sa_flags = 0;

  if (sigaction(SIGINT, &sa_ignore, &sa_intr) < 0)
    return -1;

  if (sigaction(SIGQUIT, &sa_ignore, &sa_quit) < 0)
    return -1;

  sigemptyset(&set_chldmask);
  sigaddset(&set_chldmask, SIGCHLD);

  if (sigprocmask(SIG_BLOCK, &set_chldmask, &set_save) < 0) {
    return -1;
  }

  pid = fork();
  switch (pid)
    {
      case -1: /* fork failure */
        status = -1;
	break;

      case 0: /* child proc */
	sigaction(SIGINT, &sa_intr, NULL);
	sigaction(SIGQUIT, &sa_quit, NULL);
	sigprocmask(SIG_SETMASK, &set_save, NULL);
	closeall(3);  
	PRIVS_REVOKE
        errno = 0;  
	execle("/bin/sh","sh","-c",cmd,(char*)NULL, NULL);
        exit(errno);
	break;
	    
      default: /* parent proc */
	while (waitpid(pid, &status, 0) < 0)
	if (errno != EINTR) 
	  {
	    status = -1;
	    break;
	  }
	break;
    }


  if (sigaction(SIGINT, &sa_intr, NULL) < 0) 
    return -1;
  

  if (sigaction(SIGQUIT, &sa_quit, NULL) < 0)
    return -1;

  if (sigprocmask(SIG_SETMASK, &set_save, NULL) < 0) 
    return -1;

  if (status > -1)
    return WEXITSTATUS(status);

  return status;
}


/* daemon() - detach process from user and disappear into the background */
/* returns -1 on failure, but you can't do much except exit in that case */
/* since we may already have forked. This is based on the BSD version, */
/* so the caller is responsible for things like the umask, etc. */

/* believed to work on all Posix systems */

int daemonize(void)
{

  switch (fork())
    {
    case 0:
      break;
    case -1:
      return (-1);
    default:
      _exit(0);		/* exit the original process */

    }


  if (setsid() < 0)		/* shoudn't fail */
    return (-1);


  /* dyke out this switch if you want to acquire a control tty in */
  /* the future -- not normally advisable for daemons */

  switch (fork())
    {
    case 0:
      break;
    case -1:
      return (-1);
    default:
      _exit(0);

    }

  closeall(0);

  open("/dev/null", O_RDWR);
  dup(0);
  dup(0);

  chdir("/");

  umask(022);

  return (0);

}


/* lookup UID from user */
uid_t lookupUID(char *user)
{
  struct passwd *u = NULL;

  errno = 0;
  u = getpwnam(user);
  if ((!u) && (errno != 0))
    {
      print_log(LOG_ERR, "Error looking up UID -- %s", strerror(errno));
      safe_exit(-1);
    }

  if (!u)
    {
      print_log(LOG_ERR, "User %s not found", user);
      safe_exit(-1);
    }
  return (u->pw_uid);
}

/* lookup GID from group */
gid_t lookupGID(char *group)
{
  struct group *g = NULL;

  errno = 0;
  g = getgrnam(group);
  if ((!g) && (errno != 0))
    {
      print_log(LOG_ERR, "Error looking up GID -- %s", strerror(errno));
      safe_exit(-1);
    }

  if (!g)
    {
      print_log(LOG_ERR, "Group %s not found", group);
      safe_exit(-1);
    }
  return (g->gr_gid);

}


/* parses the xferlog line */
int parseLogLine(char *input)
{
  char *tmp1, *tmp2;

  memset(ftpFilename, 0, sizeof(ftpFilename));
  memset(ftpCommand, 0, sizeof(ftpCommand));

  tmp1 = strsep(&input, " ");
  tmp2 = strsep(&input, "\n");
  if ((tmp1 != NULL) && (tmp2 != NULL))
    {
      strcpy(ftpFilename, tmp2);
      strcpy(ftpCommand, tmp1);
      return (1);
    }
  else
    return (-1);
}


/* parses the commandline */
int parseCommandline(int argc, char *argv[])
{
  int c;

  while (1)
    {

      c = getopt(argc, argv, "c:hvn");

      if (c == -1)
        return (0);

      switch (c)
        {
        case 'h':
          fprintf(stderr, "ftpexecd v%s\n", VERSION);
          fprintf(stderr, "usage:\n");
          fprintf(stderr, " -c file\tuses configfile <file>\n");
          fprintf(stderr, " -v\t\tdisplay version\n");
          fprintf(stderr, " -h\t\tdisplay help\n");
          fprintf(stderr," -n\t\tdisables daemon mode\n");
          return (-1);

        case 'v':
          fprintf(stderr, "ftpexecd v%s\n", VERSION);
          fprintf(stderr, "Written by Thomas Lachmann <TL@bunker-werk.net>\n");
          return (-1);

        case 'c':
          fprintf(stderr, "using config-file %s\n", optarg);
          configFile = optarg;
          break;

        case 'n':
          fprintf(stderr, "starting in non-daemon mode\n");
          nodaemon_flag = 1;
          break;

        }

    }
}


/* read the config file */
void read_config(void)
{
  FILE *cf;
  char line[MAX_LINE_LEN + 1];
  char *token;

  PRIVS_ROOT
  cf = fopen(configFile, "r");
  PRIVS_RELINQUISH
  if (cf == NULL)
    {
      print_log(LOG_ERR, "can't open config-file %s -- %s", configFile,
                strerror(errno));
      safe_exit(-1);
    }

  while (fgets(line, MAX_LINE_LEN, cf) != NULL)
    {
      token = strtok(line, "\t =\n\r");
      if ((token != NULL) && (token != "#"))
        {
          if (strcasecmp(token, "fifoFile") == 0)
            fifoFile = strdup(strtok(NULL, "\t =\n\r"));

          if (strcasecmp(token, "logFile") == 0)
            logFile = strdup(strtok(NULL, "\t =\n\r"));

          if (strcasecmp(token, "Exec") == 0)
            execProgram = strdup(strtok(NULL, "\t =\n\r"));

          if (strcasecmp(token, "User") == 0)
            daemonUser = strdup(strtok(NULL, "\t =\n\r"));

          if (strcasecmp(token, "Group") == 0)
            daemonGroup = strdup(strtok(NULL, "\t =\n\r"));

        }
    }
  fclose(cf);

  if (fifoFile == NULL)
    {
      print_log(LOG_ERR, "no named pipe/fifo specified");
      safe_exit(-1);
    }
}


/* checks if we have acces to the files (in-fifo, out-fifo(log) */
void update_config(void)
{
  struct stat stat_buf;
  int fd;

  if (strcasecmp(logFile, "none") != 0)
    {
      writeLog = 1;

      if (strcasecmp(logFile, "syslog") != 0)
        {
          syslog_flag = 0;

          PRIVS_ROOT
          fileLog = fopen(logFile, "a+");
          PRIVS_RELINQUISH
          if (fileLog == NULL)
            {
              syslog_flag = 1;
              print_log(LOG_ERR, "failed to open logFile %s -- %s",
                        logFile, strerror(errno));
              safe_exit(-1);
            }
          fclose(fileLog);
        }
      else
        syslog_flag = 1;

    }
  else
    writeLog = 0;

  PRIVS_ROOT
  PRIVS_SETUP(lookupUID(daemonUser), lookupGID(daemonGroup))
  
  fd = open(fifoFile, O_RDONLY | O_NONBLOCK);

  if (fd == -1)
    {
      print_log(LOG_ERR, "failed to open fifo %s -- %s", fifoFile,
                strerror(errno));
      safe_exit(-1);
    }
  close(fd);

  if (nodaemon_flag)
    {
      fprintf(stderr, "Daemon runs as\n");
      fprintf(stderr, "User = %s -- UID = %i\n", daemonUser, geteuid());
      fprintf(stderr, "Group = %s -- GID = %i\n", daemonGroup, getegid());
      fprintf(stderr, "Using fifo = %s\n", fifoFile);
      fprintf(stderr, "Using log-file = %s\n", logFile);
    }

  if (stat(fifoFile, &stat_buf) == 0)
    {
      if (!S_ISFIFO(stat_buf.st_mode))
        {
	  print_log(LOG_ERR, "%s is not a fifo", fifoFile);
	  safe_exit(-1);
	}
      if (stat_buf.st_mode & S_IWOTH)
        print_log(LOG_WARNING, "%s is world-writable", fifoFile);
    }
    else print_log(LOG_ERR, "error stat'ing %s: %s", fifoFile, strerror(errno));
}

/* signalTerm() -- handler for SIGTERM, close all files*/
void signalTerm()
{
  safe_exit(0);
}


/* signalHup() -- handler for SIGHUP, updates configuration */
void signalHup()
{
  print_log(LOG_INFO, "received SIGHUP - update config");
  read_config();
  update_config();
}

/* setup all SIG handlers we need */
void init_sigs(void)
{}

/* This is the daemon's main work -- listen for connections and spawn */
void process()
{
  char input[MAX_LINE_LEN];
  char tmp[MAX_LINE_LEN + 400];

  while (1)
    {

      fileFifo = fopen(fifoFile, "r");

      if (fileFifo != NULL)
        {
          while (fgets(input, MAX_LINE_LEN, fileFifo) != NULL)
            {
              input[strlen(input) - 1] = '\0';

              if (nodaemon_flag)
                print_log(LOG_DEBUG, "input from fifo: <%s>", input);

              if (parseLogLine(input))
                if ((strcmp(ftpCommand, "STOR") == 0) || (strcmp(ftpCommand, "APPE") == 0) || (strcmp(ftpCommand, "STOU") == 0))
                  {	/* STOR or APPE or STOU command ? */
                    sprintf(tmp, "%s \"%s\" &", execProgram,
                            ftpFilename);
                    if (file_exists(ftpFilename))
                      {
                        if (safe_exec(tmp) == -1)
                          print_log(LOG_ERR, "Failed to execute %s -- %s", tmp, strerror(errno));
                      }
                    else
                      print_log(LOG_WARNING,
                                "Uploaded file %s doesn't exist, script not executed",
                                ftpFilename);
                  }
            }
          fclose(fileFifo);
        }
      else
        {
          print_log(LOG_ERR, "failed to open fifo %s -- %s", fifoFile,
                    strerror(errno));
          safe_exit(-1);
        }
    }
}

int main(int argc, char *argv[])
{
  if (parseCommandline(argc, argv) != 0)
    return (1);

  init_sigs();

  read_config();

  update_config();

  if (nodaemon_flag == 0)
    if (daemonize() < 0)
      {
        perror("daemon error");
        safe_exit(2);
      }


  signal(SIGTERM, signalTerm);
  signal(SIGQUIT, signalTerm);
  signal(SIGINT, signalTerm);
  signal(SIGKILL, signalTerm);
  signal(SIGHUP, signalHup);
  openlog("ftpexecd", LOG_PID | LOG_NDELAY, LOG_DAEMON);
  print_log(LOG_INFO, "started ftpexecd");
  process();
  return (0);
}
